"""init file"""
from pathlib import Path
import sys
from io import IOBase
import numpy as np

from .frame_reader import FrameReader
from .pil_frame_reader import PILFrameReader
from .numpy_frame_reader import NumpyFrameReader
from .ffmpeg_frame_reader import FFmpegFrameReader
from .fd_frame_reader import FdFrameReader

def build_frame_reader(source: str | Path | FrameReader | list[np.ndarray] | np.ndarray | IOBase,
                       **kwargs) -> FrameReader:
    """builds the frame reader given a source path for a VREVideo"""
    if isinstance(source, FrameReader):
        return source
    # pylint: disable=not-an-iterable
    if isinstance(source, np.ndarray) or (isinstance(source, list) and all(isinstance(x, np.ndarray) for x in source)):
        return NumpyFrameReader(source, **kwargs)
    if isinstance(source, IOBase):
        return FdFrameReader(source, **kwargs)
    if isinstance(source, str) and source == "-":
        return FdFrameReader(sys.stdin.buffer, **kwargs)
    if Path(source).is_dir():
        suffixes = list({x.suffix for x in Path(source).iterdir()})
        assert len(suffixes) == 1, suffixes
        if suffixes[0] == ".png":
            return PILFrameReader(source, **kwargs)
        if suffixes[0] in (".npy", ".npz"):
            return NumpyFrameReader(source, **kwargs)
    # Otherwise, let ffmpeg handle it and it'll throw on errors.
    return FFmpegFrameReader(source, **kwargs)



"""init file"""
import os
from .frame_writer import FrameWriter
from .pil_frame_writer import PILFrameWriter
from .numpy_frame_writer import NumpyFrameWriter
from .ffmpeg_frame_writer import FFmpegFrameWriter
from ..utils import logger

def build_frame_writer(writer: str | FrameWriter | None, **kwargs) -> FrameWriter:
    """builds the frame writer given a source path for a VREVideo"""
    if isinstance(writer, FrameWriter):
        return writer
    if writer is None:
        writer = os.getenv("VRE_VIDEO_DEFAULT_WRITER", "pillow")
        logger.debug(f"No writer provided. Defaulting to: '{writer}'. Change with 'VRE_VIDEO_DEFAULT_WRITER' env var.")
    if writer in ("pillow", "PIL"):
        return PILFrameWriter(**kwargs)
    if writer in ("numpy", "np"):
        return NumpyFrameWriter(**kwargs)
    if writer == "ffmpeg":
        return FFmpegFrameWriter(**kwargs)
    raise NotImplementedError(writer)



"""
fd_frame_reader - module for reading from a file descriptor (like stdin). To be used with external generating data
tool like ffmpeg (i.e. for webcams) and linux pipes.
"""
from io import IOBase
from threading import Lock, Thread
import numpy as np
from .frame_reader import FrameReader

class FdFrameReader(FrameReader):
    """Implements FrameReader that reads from an io stream (like stdin). Can be used with linux pipes."""
    def __init__(self, data: IOBase, resolution: tuple[int, int]=(360, 1280), fps: int=30):
        assert isinstance(data, IOBase), type(data)
        super().__init__()
        self.data = data
        self._fps = fps
        self.resolution = (self.height, self.width) = resolution

        self.lock = Lock()
        self.current_frame: np.ndarray | None = np.zeros((*resolution, 3), dtype=np.uint8)
        Thread(target=self._read_frame_worker, daemon=True).start()

    @property
    def shape(self) -> tuple[int, int, int, int]:
        return (1, self.height, self.width, 3)

    @property
    def fps(self):
        return self._fps

    @property
    def path(self) -> str:
        return str(self.data)

    def _read_frame_worker(self) -> np.ndarray:
        """note: we need to use a thread here otherwise ffmpeg is too fast for our sync reader"""
        while True:
            raw_frame = self.data.read(self.width * self.height * 3)
            frame = None
            if raw_frame is not None:
                frame = np.frombuffer(raw_frame, dtype=np.uint8).reshape((self.height, self.width, 3))
            with self.lock:
                self.current_frame = frame

    def __getitem__(self, ix: int | list[int] | np.ndarray | slice) -> np.ndarray:
        with self.lock:
            frame = self.current_frame
        if frame is None:
            raise StopIteration
        if isinstance(ix, slice):
            assert (ix.start - ix.stop) == 1, f"cannot have batches in fd reader, got: {list(ix)}"
            ix = [0]
        if isinstance(ix, (list, np.ndarray)):
            assert len(ix) == 1, f"cannot have batches in fd reader, got: {ix}"
            frame = frame[None] # return a batch of one for other tools
        return frame



"""ffmpeg_frame_writer - Module that implements frame writing using ffmpeg"""
from pathlib import Path
from io import FileIO
import os
import subprocess
from tqdm import trange
import numpy as np

from .frame_writer import FrameWriter
from ..utils import logger

class FFmpegFrameWriter(FrameWriter):
    """FFmpegFrameWriter implementation that writes to disk a video using ffmpeg process"""
    def write(self, video: "VREVideo", out_path: Path, start_frame: int = 0, end_frame: int | None = None):
        out_path = Path(out_path)
        assert out_path.suffix == ".mp4", out_path
        assert isinstance(start_frame, int) and start_frame >= 0, start_frame

        process, logfile = self._start_ffmpeg_write_process(video, out_path)

        assert not out_path.exists(), out_path
        out_path.parent.mkdir(exist_ok=True, parents=True)
        start_frame = start_frame or 0
        end_frame = end_frame or len(video)
        assert start_frame >= 0 and end_frame <= len(video), (start_frame, end_frame)

        try:
            for i in trange(start_frame, end_frame, disable=os.getenv("VRE_VIDEO_PBAR", "1") == "0",
                            desc=f"[FFmpegFrameWriter] {out_path}"):
                frame: np.ndarray = video[i]
                process.stdin.write(frame.tobytes())
        finally:
            process.stdin.close()
            process.wait()
            logfile.close()

    def _start_ffmpeg_write_process(self, video: "VREVideo", out_path: Path) -> tuple[subprocess.Popen, FileIO]:
        log_file = open(out_path.with_suffix(".ffmpeg.log"), "w")
        cmd = [
            "ffmpeg",
            "-y",  # overwrite output file if it exists
            "-f", "rawvideo",
            "-vcodec", "rawvideo",
            "-pix_fmt", "rgb24",
            "-s", f"{video.shape[2]}x{video.shape[1]}",
            "-r", str(video.fps),
            "-i", "pipe:0",             # input comes from stdin
            "-an",                      # no audio
            "-vcodec", "libx264",
            "-pix_fmt", "yuv420p",
            str(out_path)
        ]
        logger.debug(f"Running '{' '.join(cmd)}' (log: {log_file.name})")
        process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.DEVNULL, stderr=log_file)
        return process, log_file



"""VREVideo module"""
from __future__ import annotations
from pathlib import Path
from typing import Iterable
from io import FileIO
import numpy as np

from .readers import FrameReader, build_frame_reader
from .writers import FrameWriter, build_frame_writer

class VREVideo(Iterable):
    """VREVideo -- A generic wrapper on top of a Video container"""
    def __init__(self, source: str | Path | FrameReader | np.ndarray | list[np.ndarray] | FileIO, **reader_kwargs):
        self.reader: FrameReader = build_frame_reader(source, **(reader_kwargs or {}))

    @property
    def frame_shape(self) -> tuple[int, int, int]:
        """the shape of a frame"""
        return self.shape[1:]

    @property
    def shape(self) -> tuple[int, int, int, int]:
        """The shape of this video"""
        return self.reader.shape

    @property
    def fps(self) -> float:
        """The FPS of this video"""
        return self.reader.fps

    @property
    def path(self) -> str:
        """The path of the video"""
        return self.reader.path

    def write(self, out_path: Path, start_frame: int = 0, end_frame: int | None = None, **writer_kwargs) -> FrameWriter:
        """writes this video to the disk"""
        if Path(out_path).suffix != "": # .mp4 etc.
            writer = build_frame_writer("ffmpeg")
        else:
            writer = build_frame_writer(**writer_kwargs)
        writer.write(self, out_path, start_frame, end_frame)
        return writer

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, ix: int | list[int] | np.ndarray | slice) -> np.ndarray:
        assert isinstance(ix, (int, list, np.ndarray, slice)), type(ix)
        return self.reader[ix]

    def __len__(self):
        return len(self.reader)

    def __repr__(self):
        return f"[VREVideo] {self.reader}"



"""frame_writer module. Other readers will implement these methods."""
from abc import ABC, abstractmethod
from pathlib import Path

class FrameWriter(ABC):
    """FrameWriter - defines the interface of a reader"""
    @abstractmethod
    def write(self, video: "VREVideo", out_path: str | Path, start_frame: int = 0, end_frame: int | None = None):
        """writes video frames to the disk"""



"""numpy_frame_reader - module for NumPy-based functionality"""
from pathlib import Path
from natsort import natsorted
import numpy as np
from .frame_reader import FrameReader

class NumpyFrameReader(FrameReader):
    """Implements FrameReader using NumPy arrays, from memory or from disk (.npy/.npz files)"""
    def __init__(self, data: list[np.ndarray] | np.ndarray | str | Path, fps: int = 1, frames: list[int] | None = None):
        self.data, self.frames = self._build_data(data, frames)
        self._fps = fps
        self.frames = self.frames or list(range(len(self.data)))
        self.frames_ix = dict(zip(self.frames, range(len(self.frames)))) # {ix: frame}
        self._path = str(data) if isinstance(data, (str, Path)) else "in memory"

    @property
    def shape(self) -> tuple[int, int, int, int]:
        return self.data.shape

    @property
    def fps(self):
        return self._fps

    @property
    def path(self) -> str:
        return str(self._path)

    def _build_data(self, data: list[np.ndarray] | np.ndarray | str | Path,
                    frames: list[int] | None = None) -> tuple[np.ndarray, list[int]]:
        if isinstance(data, np.ndarray):
            assert data.ndim == 4, f"Expected 4D NumPy array (N, H, W, C), got shape {data.shape}"
            assert data.dtype == np.uint8, f"Expected dtype uint8, got {data.dtype}"
            return data, frames

        if isinstance(data, list):
            _shp = data[0].shape
            assert all(x.shape == _shp for x in data), f"Not all shapes match first image: {_shp}"
            assert all(x.dtype == np.uint8 for x in data), f"Not all data is uint8: {set(x.dtype for x in data)}"
            return np.array(data), frames

        if isinstance(data, (str, Path)):
            path = Path(data)
            assert path.exists() and path.is_dir(), f"Invalid path: '{path}'"

            files = natsorted([p for p in path.iterdir() if p.suffix in {".npy", ".npz"}], key=lambda p: p.name)
            suffixes = {f.suffix for f in files}
            assert len(suffixes) == 1, f"Directory must contain only one type of file (.npy or .npz), found: {suffixes}"

            if files[0].suffix == ".npy":
                data_lst: list[np.ndarray] = [np.load(f) for f in files]
            else:
                data_lst: list[np.ndarray] = [np.load(f)["arr_0"] for f in files]

            # try to make the frames based on 1.npz, 2.npz, etc.
            if frames is None:
                try:
                    frames = sorted([int(x.stem) for x in files])
                except Exception:
                    pass

            assert all(x.shape == data_lst[0].shape for x in data_lst), f"Shapes differ: expected {data_lst[0].shape}"
            assert all(x.dtype == np.uint8 for x in data_lst), f"Dtype mismatch: {[x.dtype for x in data_lst]}"
            return np.array(data_lst), frames

        raise TypeError(f"Data must be a list of np.ndarray, a 4D np.ndarray, or a directory path: Got {type(data)}.")

    def __getitem__(self, ix: int | list[int] | np.ndarray | slice) -> np.ndarray:
        if isinstance(ix, int):
            return self.data[self.frames_ix[ix]]
        if isinstance(ix, np.ndarray):
            ix = ix.tolist()
        if isinstance(ix, slice):
            ix = list(range(ix.start, ix.stop))
        ix_transformed = [self.frames_ix[_ix] for _ix in ix] # support for sparse frames (VRE)
        return self.data[ix_transformed]

    def __len__(self):
        return self.frames[-1] + 1



"""FFmpegFrameReader module"""
from pathlib import Path
from io import FileIO
import json
import subprocess
import numpy as np

from .frame_reader import FrameReader
from ..utils import logger

class FFmpegFrameReader(FrameReader):
    """FFmpegFrameReader -- reads data from a video using ffmpeg"""
    def __init__(self, path: str | Path, cache_len: int = 100):
        super().__init__()
        self._path = self._build_path(path)
        assert self._path.exists(), f"Video '{self._path}' doesn't exist"
        self.probe_data, self.stream_info = self._run_ffprobe_get_data()

        self._fps: float | None = None
        self._shape: tuple[int, int, int, int] | None = None

        self.cache = []
        self.cache_max_len = cache_len
        self.cache_start_frame = None
        self.cache_end_frame = None

        self.process: subprocess.Popen | None = None
        self.log_file: FileIO | None = None

    @property
    def shape(self) -> tuple[int, int, int, int]:
        if self._shape is None:
            width = int(self.stream_info["width"])
            height = int(self.stream_info["height"])
            total_frames = self._build_total_frames()
            self._shape = (total_frames, height, width, 3)
        return self._shape

    @property
    def fps(self) -> float:
        if self._fps is None:
            self._fps = eval(self.stream_info["avg_frame_rate"]) # pylint: disable=eval-used
        return self._fps

    @property
    def path(self) -> str:
        return str(self._path)

    def get_one_frame(self, ix: int) -> np.ndarray:
        """Retrieve a frame from the video by frame number, using nearby frames caching."""
        assert isinstance(ix, int), type(ix)
        assert 0 <= ix < len(self), f"Frame out of bounds: {ix}. Len: {len(self)}"

        # Load new cache if the requested frame is outside the current cache range
        if self.cache_start_frame is None or not self.cache_start_frame <= ix < self.cache_end_frame:
            self._cache_frames(start_frame=ix)

        # Calculate the index within the cache
        return self.cache[ix - self.cache_start_frame]

    def _build_path(self, path: str | Path) -> Path:
        """Builds the path. Can also be a youtube video, not just a local path, but yt_dlp must be installed"""
        if (s_path := str(path)).startswith("http") and (s_path.find("youtube") != -1 or s_path.find("youtu.be") != -1):
            from yt_dlp import YoutubeDL # pylint: disable=import-outside-toplevel, import-error
            tmpfile = f"/tmp/{path}.mp4"
            if not Path(tmpfile).exists():
                with YoutubeDL({"format": "bv*", "outtmpl": tmpfile, "quiet": True,
                                "no_warnings": True, "noprogress": True}) as ydl:
                    ydl.download([s_path])
            path = tmpfile
        return Path(path)

    def _build_total_frames(self) -> int:
        """returns the number of frames of the vifdeo"""
        if "nb_frames" in self.stream_info:
            return int(float(self.stream_info["nb_frames"]))
        if "duration" in self.stream_info:
            duration_s = float(self.stream_info["duration"])
            return int(duration_s * self.fps)
        if self.stream_info.get("codec_name", "") == "h264":
            if "tags" in self.stream_info and "DURATION" in self.stream_info["tags"]:
                duration_str = self.stream_info["tags"]["DURATION"]
                h, m, s = [float(x) for x in duration_str.split(":")]
                duration_s = h * 60 * 60 + m * 60 + s
                return int(duration_s * self.fps)
        if self.stream_info.get("codec_name", "") == "vp9":
            if "format" in self.probe_data and "duration" in self.probe_data["format"]:
                duration_s = float(self.probe_data["format"]["duration"])
                return int(duration_s * self.fps)
        raise ValueError(f"Unknown video format. Stream info from ffmpeg: {self.stream_info}")

    def _start_ffmpeg_process(self, start_time: float) -> subprocess.Popen:
        """
        Start an ffmpeg process from the nearest keyframe to the requested start_time.
        This will load all frames from this keyframe to the next keyframe.
        """
        self._cleanup_if_needed()
        self.log_file = open(Path(self.path).with_suffix(".ffmpeg.log"), "w")
        cmd = [
            "ffmpeg",
            "-ss", str(start_time),
            "-i", str(self.path),
            "-f", "rawvideo",
            "-pix_fmt", "rgb24",
            "pipe:" # output to stdout
        ]
        logger.debug(f"Running '{' '.join(cmd)}' (log: {self.log_file.name})")
        self.process = subprocess.Popen(cmd, stdin=subprocess.DEVNULL, stdout=subprocess.PIPE, stderr=self.log_file)

    def _run_ffprobe_get_data(self) -> tuple[dict, dict]:
        cmd = [
            "ffprobe",
            "-v", "error",            # suppress unnecessary output
            "-select_streams", "v",   # select only video streams
            "-show_entries", "stream",  # show all stream fields
            "-show_entries", "format", # required for vp9
            "-of", "json",            # output format: JSON
            str(self.path)
        ]
        logger.debug(f"Running '{' '.join(cmd)}'")
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        probe_data = json.loads(result.stdout)

        # Extract the first video stream
        stream_info = next((stream for stream in probe_data.get("streams", [])
                            if stream.get("codec_type") == "video"), None)
        return probe_data, stream_info

    def _cleanup_if_needed(self):
        if hasattr(self, "process") and self.process is not None: # hasattr in case the constructor throws on .exists()
            self.process.stdout.close()
            self.process.terminate()
        if hasattr(self, "log_file") and self.log_file is not None:
            self.log_file.close()
            self.log_file = None

    def _cache_frames(self, start_frame: int):
        """Cache all frames between the current keyframe and the next keyframe, starting at start_frame."""
        start_time = start_frame / self.fps
        self.cache = []
        if self.process is None or start_frame != self.cache_end_frame:
            self._start_ffmpeg_process(start_time)

        # Read frames until the end of the current keyframe range
        while True:
            in_bytes = self.process.stdout.read(self.shape[1] * self.shape[2] * 3)
            if not in_bytes or len(self.cache) > self.cache_max_len:
                break
            frame = np.frombuffer(in_bytes, np.uint8).reshape([self.shape[1], self.shape[2], 3])
            self.cache.append(frame)

        # Set cache boundaries in terms of frame numbers
        self.cache_start_frame = start_frame
        self.cache_end_frame = start_frame + len(self.cache)

    def __del__(self):
        self._cleanup_if_needed()

    def __getitem__(self, ix: int | list[int] | np.ndarray | slice) -> np.ndarray:
        if isinstance(ix, np.ndarray):
            return self[ix.tolist()]
        if isinstance(ix, list):
            return np.array([self[_ix] for _ix in ix])
        if isinstance(ix, slice):
            return np.array([self[_ix] for _ix in range(ix.start, ix.stop)])
        return self.get_one_frame(ix)



"""setup.py -- note use setuptools==73.0.1; older versions fuck up the data files, newer versions include resources."""
from pathlib import Path
from setuptools import setup, find_packages

NAME = "vre-video"
VERSION = "0.3.3"
DESCRIPTION = "Video Representations Extractor (VRE) Video reader"
URL = "https://gitlab.com/video-representations-extractor/vre-video"

CWD = Path(__file__).absolute().parent
with open(CWD/"README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

REQUIRED = [
    "numpy>=1.21.6,<2.0.0",
    "tqdm==4.66.5",
    "Pillow==10.3.0",
    "natsort==8.2.0",
    "loggez>=0.4.4",
]

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    packages=find_packages(),
    install_requires=REQUIRED,
    dependency_links=[],
    license="MIT",
    python_requires=">=3.8",
)
